@@ -1,7 +1,7 @@
 # vim: ts=8:sts=8:sw=8:noexpandtab
 
 # This file is part of ReText
-# Copyright: 2012-2017 Dmitry Shachnev
+# Copyright: 2012-2020 Dmitry Shachnev
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@ from os.path import dirname, exists, join
 from PyQt5.QtCore import QByteArray, QLocale, QSettings, QStandardPaths
 from PyQt5.QtGui import QFont
 
-app_version = "7.0.3"
+app_version = "7.1.0"
 
 settings = QSettings('ReText project', 'ReText')
 
@@ -38,8 +38,10 @@ datadirs = []
 def initializeDataDirs():
 	assert not datadirs
 
-	if '__file__' in locals():
+	try:
 		datadirs.append(dirname(dirname(__file__)))
+	except NameError:
+		pass
 
 	dataLocations = QStandardPaths.standardLocations(QStandardPaths.GenericDataLocation)
 	datadirs.extend(join(d, 'retext') for d in dataLocations)
@@ -67,7 +69,9 @@ configOptions = {
 	'autoSave': False,
 	'defaultCodec': '',
 	'defaultMarkup': markups.MarkdownMarkup.name,
+	'defaultPreviewState': 'editor',
 	'detectEncoding': True,
+	'documentStatsEnabled': False,
 	'editorFont': QFont(),
 	'font': QFont(),
 	'handleWebLinks': False,
@@ -76,12 +80,16 @@ configOptions = {
 	'iconTheme': '',
 	'lastTabIndex': 0,
 	'lineNumbersEnabled': False,
-	'livePreviewByDefault': False,
 	'markdownDefaultFileExtension': '.mkd',
+	'openFilesInExistingWindow': True,
 	'openLastFilesOnStartup': False,
+	'paperSize': '',
 	'pygmentsStyle': 'default',
+	'recentDocumentsCount': 10,
+	'relativeLineNumbers': False,
 	'restDefaultFileExtension': '.rst',
 	'rightMargin': 0,
+	'rightMarginWrap': False,
 	'saveWindowGeometry': False,
 	'spellCheck': False,
 	'spellCheckLocale': '',
@@ -94,7 +102,9 @@ configOptions = {
 	'useFakeVim': False,
 	'useWebEngine': False,
 	'useWebKit': False,
+	'wideCursor': False,
 	'windowGeometry': QByteArray(),
+	'windowTitleFullPath': False,
 }
 
 def readFromSettings(key, keytype, settings=settings, default=None):
@@ -141,6 +151,9 @@ def writeListToSettings(key, value, settings=settings):
 	else:
 		settings.remove(key)
 
+def getSettingsFilePath(settings=settings):
+	return settings.fileName()
+
 def chooseMonospaceFont():
 	font = QFont('monospace')
 	font.setStyleHint(QFont.TypeWriter)
