@@ -1,7 +1,7 @@
 # vim: ts=8:sts=8:sw=8:noexpandtab
 #
 # This file is part of ReText
-# Copyright: 2015-2017 Dmitry Shachnev
+# Copyright: 2015-2020 Dmitry Shachnev
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -31,8 +31,8 @@ except ImportError:
 	enchant = None
 
 from PyQt5.QtCore import pyqtSignal, Qt, QDir, QFile, QFileInfo, QPoint, QTextStream, QTimer, QUrl
-from PyQt5.QtGui import QTextCursor, QTextDocument
-from PyQt5.QtWidgets import QTextEdit, QSplitter
+from PyQt5.QtGui import QPalette, QTextCursor, QTextDocument
+from PyQt5.QtWidgets import QApplication, QTextEdit, QSplitter, QMessageBox
 
 try:
 	from ReText.webkitpreview import ReTextWebKitPreview
@@ -61,7 +61,7 @@ class ReTextTab(QSplitter):
 		return self._fileName
 
 	def __init__(self, parent, fileName, previewState=PreviewDisabled):
-		super(QSplitter, self).__init__(Qt.Horizontal, parent=parent)
+		super().__init__(Qt.Horizontal, parent=parent)
 		self.p = parent
 		self._fileName = fileName
 		self.editBox = ReTextEdit(self)
@@ -72,6 +72,7 @@ class ReTextTab(QSplitter):
 		self.previewState = previewState
 		self.previewOutdated = False
 		self.conversionPending = False
+		self.cssFileExists = False
 
 		self.converterProcess = converterprocess.ConverterProcess()
 		self.converterProcess.conversionDone.connect(self.updatePreviewBox)
@@ -195,19 +196,30 @@ class ReTextTab(QSplitter):
 				errMsg = errMsg.replace('<a href="%s">', '').replace('</a>', '')
 			return '<p style="color: red">%s</p>' % errMsg
 		headers = ''
-		if includeStyleSheet:
+		if includeStyleSheet and self.p.ss is not None:
 			headers += '<style type="text/css">\n' + self.p.ss + '</style>\n'
+		elif includeStyleSheet:
+			style = 'td, th { border: 1px solid #c3c3c3; padding: 0 3px 0 3px; }\n'
+			style += 'table { border-collapse: collapse; }\n'
+			style += 'img { max-width: 100%; }\n'
+			# QTextDocument seems to use media=screen even for printing
+			if globalSettings.useWebKit:
+				# https://github.com/retext-project/retext/pull/187
+				palette = QApplication.palette()
+				style += '@media screen { html { color: %s; } }\n' % palette.color(QPalette.WindowText).name()
+				# https://github.com/retext-project/retext/issues/408
+				style += '@media print { html { background-color: white; } }\n'
+			headers += '<style type="text/css">\n' + style + '</style>\n'
 		baseName = self.getBaseName()
-		cssFileName = baseName + '.css'
-		if QFile.exists(cssFileName):
-			headers += ('<link rel="stylesheet" type="text/css" href="%s">\n'
-			% cssFileName)
+		if self.cssFileExists:
+			headers += ('<link rel="stylesheet" type="text/css" href="%s.css">\n'
+			% baseName)
 		headers += ('<meta name="generator" content="ReText %s">\n' % app_version)
 		return converted.get_whole_html(
 			custom_headers=headers, include_stylesheet=includeStyleSheet,
 			fallback_title=baseName, webenv=webenv)
 
-	def getDocumentForExport(self, includeStyleSheet, webenv):
+	def getDocumentForExport(self, includeStyleSheet=True, webenv=False):
 		markupClass = self.getActiveMarkupClass()
 		if markupClass and markupClass.available():
 			exportMarkup = markupClass(filename=self._fileName)
@@ -274,8 +286,9 @@ class ReTextTab(QSplitter):
 			requested_extensions = ['ReText.mdx_posmap'] if globalSettings.syncScroll else []
 			self.converterProcess.start_conversion(self.getActiveMarkupClass().name,
 			                                       self.fileName,
-							       requested_extensions,
-							       self.editBox.toPlainText())
+			                                       requested_extensions,
+			                                       self.editBox.toPlainText(),
+			                                       QDir.currentPath())
 
 	def updateBoxesVisibility(self):
 		self.editBox.setVisible(self.previewState < PreviewNormal)
@@ -345,12 +358,15 @@ class ReTextTab(QSplitter):
 		self.editBox.setPlainText(text)
 		self.editBox.document().setModified(False)
 
+		cssFileName = self.getBaseName() + '.css'
+		self.cssFileExists = QFile.exists(cssFileName)
+
 		if previousFileName != self._fileName:
 			self.fileNameChanged.emit()
 
 	def writeTextToFile(self, fileName=None):
 		# Just writes the text to file, without any changes to tab object
-		# Used directly for i.e. export extensions
+		# Used directly for e.g. export extensions
 
 		# Get text from the cursor to avoid tweaking special characters,
 		# see https://bugreports.qt.io/browse/QTBUG-57552 and
@@ -388,6 +404,12 @@ class ReTextTab(QSplitter):
 
 		return result
 
+	def goToLine(self,line):
+		block = self.editBox.document().findBlockByLineNumber(line)
+		if block.isValid():
+			newCursor = QTextCursor(block)
+			self.editBox.setTextCursor(newCursor)
+
 	def find(self, text, flags, replaceText=None, wrap=False):
 		cursor = self.editBox.textCursor()
 		if wrap and flags & QTextDocument.FindBackward:
@@ -404,6 +426,10 @@ class ReTextTab(QSplitter):
 				newCursor.movePosition(QTextCursor.Left, QTextCursor.MoveAnchor, len(replaceText))
 				newCursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, len(replaceText))
 			self.editBox.setTextCursor(newCursor)
+			if self.editBox.cursorRect().bottom() >= self.editBox.height() - 3:
+				scrollValue = self.editBox.verticalScrollBar().value()
+				areaHeight = self.editBox.fontMetrics().height()
+				self.editBox.verticalScrollBar().setValue(scrollValue + areaHeight)
 			return True
 		if not wrap:
 			return self.find(text, flags, replaceText, True)
@@ -426,19 +452,62 @@ class ReTextTab(QSplitter):
 		self.editBox.textCursor().endEditBlock()
 		return not lastCursor.isNull()
 
-	def openSourceFile(self, fileToOpen):
+	def openSourceFile(self, linkPath):
 		"""Finds and opens the source file for link target fileToOpen.
 
 		When links like [test](test) are clicked, the file test.md is opened.
 		It has to be located next to the current opened file.
 		Relative paths like [test](../test) or [test](folder/test) are also possible.
 		"""
-		if self.fileName:
-			currentExt = splitext(self.fileName)[1]
-			basename, ext = splitext(fileToOpen)
-			if ext in ('.html', '') and exists(basename + currentExt):
-				self.p.openFileWrapper(basename + currentExt)
-				return basename + currentExt
+
+		fileToOpen = self.resolveSourceFile(linkPath)
 		if exists(fileToOpen) and get_markup_for_file_name(fileToOpen, return_class=True):
 			self.p.openFileWrapper(fileToOpen)
 			return fileToOpen
+		if get_markup_for_file_name(fileToOpen, return_class=True):
+			if not QFile.exists(fileToOpen) and QFileInfo(fileToOpen).dir().exists():
+				if self.promptFileCreation(fileToOpen):
+					self.p.openFileWrapper(fileToOpen)
+					return fileToOpen
+
+	def promptFileCreation(self, fileToCreate):
+		"""
+		Prompt user if a file should be created for the clicked link,
+		and try to create it. Return True on success.
+		"""
+		buttonReply = QMessageBox.question(self, self.tr('Create missing file?'),
+		                                   self.tr("The file '%s' does not exist.\n\nDo you want to create it?") % fileToCreate,
+		                                   QMessageBox.Yes | QMessageBox.No,
+		                                   QMessageBox.No)
+		if buttonReply == QMessageBox.Yes:
+			return self.createFile(fileToCreate)
+		elif buttonReply == QMessageBox.No:
+			return False
+
+	def resolveSourceFile(self, linkPath):
+		"""
+		Finds the actual path of the file to open in a new tab.
+		When the link has no extension, eg: [Test](test), the extension of the current file is assumed
+		(eg test.md for a markdown file).
+		When the link is an html file eg: [Test](test.html), the extension of the current file is assumed
+		(eg test.md for a markdown file).
+		Relative paths like [test](../test) or [test](folder/test) are also possible.
+		"""
+		basename, ext = splitext(linkPath)
+		if self.fileName:
+			currentExt = splitext(self.fileName)[1]
+			if ext in ('.html', '') and (exists(basename+currentExt) or not exists(linkPath)):
+				ext = currentExt
+
+		return basename+ext
+
+	def createFile(self, fileToCreate):
+		"""Try to create file, return True if successful"""
+		try:
+			# Create file:
+			open(fileToCreate, 'x').close()
+			return True
+		except OSError as err:
+			QMessageBox.warning(self, self.tr("File could not be created"),
+			                    self.tr("Could not create file '%s': %s") % (fileToCreate, err))
+			return False
