@@ -1,7 +1,7 @@
 # vim: ts=8:sts=8:sw=8:noexpandtab
 #
 # This file is part of ReText
-# Copyright: 2012-2015 Dmitry Shachnev
+# Copyright: 2012-2020 Dmitry Shachnev
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,12 +19,15 @@
 import markups
 import sys
 from subprocess import Popen
-from ReText import icon_path, app_version, globalSettings, readListFromSettings, \
- writeListToSettings, writeToSettings, datadirs
-from ReText.tab import ReTextTab, ReTextWebPreview, PreviewNormal, PreviewLive
+import warnings
+
+from ReText import (getBundledIcon, app_version, globalSettings,
+                    readListFromSettings, writeListToSettings, datadirs)
+from ReText.tab import (ReTextTab, ReTextWebKitPreview, ReTextWebEnginePreview,
+                        PreviewDisabled, PreviewNormal, PreviewLive)
 from ReText.dialogs import HtmlDialog, LocaleDialog
-from ReText.config import ConfigDialog
-from ReText.icontheme import get_icon_theme
+from ReText.config import ConfigDialog, setIconThemeFromSettings
+from ReText.tabledialog import InsertTableDialog
 
 try:
 	from ReText.fakevimeditor import ReTextFakeVimHandler, FakeVimMode
@@ -37,9 +40,10 @@ except ImportError:
 	enchant = None
 
 from PyQt5.QtCore import QDir, QFile, QFileInfo, QFileSystemWatcher, \
- QIODevice, QLocale, QTextCodec, QTextStream, QTimer, QUrl, Qt
+ QIODevice, QLocale, QMarginsF, QTextCodec, QTextStream, QTimer, QUrl, Qt, pyqtSlot
 from PyQt5.QtGui import QColor, QDesktopServices, QIcon, \
- QKeySequence, QPalette, QTextDocument, QTextDocumentWriter
+ QKeySequence, QPageLayout, QPageSize, QPagedPaintDevice, QPalette, \
+ QTextDocument, QTextDocumentWriter
 from PyQt5.QtWidgets import QAction, QActionGroup, QApplication, QCheckBox, \
  QComboBox, QDesktopWidget, QDialog, QFileDialog, QFontDialog, QInputDialog, \
  QLineEdit, QMainWindow, QMenu, QMessageBox, QTabWidget, QToolBar
@@ -54,20 +58,17 @@ class ReTextWindow(QMainWindow):
 			self.restoreGeometry(globalSettings.windowGeometry)
 		else:
 			self.move((screenRect.width()-self.width())/2, (screenRect.height()-self.height())/2)
-		if not screenRect.contains(self.geometry()):
-			self.showMaximized()
+			if not screenRect.contains(self.geometry()):
+				self.showMaximized()
 		if sys.platform.startswith('darwin'):
 			# https://github.com/retext-project/retext/issues/198
 			searchPaths = QIcon.themeSearchPaths()
 			searchPaths.append('/opt/local/share/icons')
+			searchPaths.append('/usr/local/share/icons')
 			QIcon.setThemeSearchPaths(searchPaths)
-		if globalSettings.iconTheme:
-			QIcon.setThemeName(globalSettings.iconTheme)
-		if QIcon.themeName() in ('hicolor', ''):
-			if not QFile.exists(icon_path + 'document-new.png'):
-				QIcon.setThemeName(get_icon_theme())
-		if QFile.exists(icon_path+'retext.png'):
-			self.setWindowIcon(QIcon(icon_path+'retext.png'))
+		setIconThemeFromSettings()
+		if QFile.exists(getBundledIcon('retext')):
+			self.setWindowIcon(QIcon(getBundledIcon('retext')))
 		elif QFile.exists('/usr/share/pixmaps/retext.png'):
 			self.setWindowIcon(QIcon('/usr/share/pixmaps/retext.png'))
 		else:
@@ -78,13 +79,13 @@ class ReTextWindow(QMainWindow):
 		self.setCentralWidget(self.tabWidget)
 		self.tabWidget.currentChanged.connect(self.changeIndex)
 		self.tabWidget.tabCloseRequested.connect(self.closeTab)
-		toolBar = QToolBar(self.tr('File toolbar'), self)
-		self.addToolBar(Qt.TopToolBarArea, toolBar)
+		self.toolBar = QToolBar(self.tr('File toolbar'), self)
+		self.addToolBar(Qt.TopToolBarArea, self.toolBar)
 		self.editBar = QToolBar(self.tr('Edit toolbar'), self)
 		self.addToolBar(Qt.TopToolBarArea, self.editBar)
 		self.searchBar = QToolBar(self.tr('Search toolbar'), self)
 		self.addToolBar(Qt.BottomToolBarArea, self.searchBar)
-		toolBar.setVisible(not globalSettings.hideToolBar)
+		self.toolBar.setVisible(not globalSettings.hideToolBar)
 		self.editBar.setVisible(not globalSettings.hideToolBar)
 		self.actionNew = self.act(self.tr('New'), 'document-new',
 			self.createNew, shct=QKeySequence.New)
@@ -108,6 +109,8 @@ class ReTextWindow(QMainWindow):
 			lambda: self.switchTab(1), shct=Qt.CTRL+Qt.Key_PageDown)
 		self.actionPrevTab = self.act(self.tr('Previous tab'), 'go-previous',
 			lambda: self.switchTab(-1), shct=Qt.CTRL+Qt.Key_PageUp)
+		self.actionCloseCurrentTab = self.act(self.tr('Close tab'), 'window-close',
+			lambda: self.closeTab(self.ind), shct=QKeySequence.Close)
 		self.actionPrint = self.act(self.tr('Print'), 'document-print',
 			self.printFile, shct=QKeySequence.Print)
 		self.actionPrint.setPriority(QAction.LowPriority)
@@ -118,9 +121,10 @@ class ReTextWindow(QMainWindow):
 			trig=self.changeEditorFont)
 		self.actionChangePreviewFont = self.act(self.tr('Change preview font'),
 			trig=self.changePreviewFont)
-		self.actionSearch = self.act(self.tr('Find text'), 'edit-find', shct=QKeySequence.Find)
-		self.actionSearch.setCheckable(True)
-		self.actionSearch.triggered[bool].connect(self.searchBar.setVisible)
+		self.actionSearch = self.act(self.tr('Find text'), 'edit-find',
+			self.search, shct=QKeySequence.Find)
+		self.actionGoToLine = self.act(self.tr('Go to line'),
+			trig=self.goToLine, shct=Qt.CTRL+Qt.Key_G)
 		self.searchBar.visibilityChanged.connect(self.searchBarVisibilityChanged)
 		self.actionPreview = self.act(self.tr('Preview'), shct=Qt.CTRL+Qt.Key_E,
 			trigbool=self.preview)
@@ -131,15 +135,19 @@ class ReTextWindow(QMainWindow):
 		elif QIcon.hasThemeIcon('x-office-document'):
 			self.actionPreview.setIcon(QIcon.fromTheme('x-office-document'))
 		else:
-			self.actionPreview.setIcon(QIcon(icon_path+'document-preview.png'))
+			self.actionPreview.setIcon(QIcon(getBundledIcon('document-preview')))
 		self.actionLivePreview = self.act(self.tr('Live preview'), shct=Qt.CTRL+Qt.Key_L,
 		trigbool=self.enableLivePreview)
 		menuPreview = QMenu()
 		menuPreview.addAction(self.actionLivePreview)
 		self.actionPreview.setMenu(menuPreview)
+		self.actionInsertTable = self.act(self.tr('Insert table'),
+			trig=lambda: self.insertFormatting('table'))
 		self.actionTableMode = self.act(self.tr('Table editing mode'),
 			shct=Qt.CTRL+Qt.Key_T,
 			trigbool=lambda x: self.currentTab.editBox.enableTableMode(x))
+		self.actionInsertImages = self.act(self.tr('Insert images by file path'),
+			trig=lambda: self.insertImages())
 		if ReTextFakeVimHandler:
 			self.actionFakeVimMode = self.act(self.tr('FakeVim mode'),
 				shct=Qt.CTRL+Qt.ALT+Qt.Key_V, trigbool=self.enableFakeVimMode)
@@ -148,6 +156,7 @@ class ReTextWindow(QMainWindow):
 				self.enableFakeVimMode(True)
 		self.actionFullScreen = self.act(self.tr('Fullscreen mode'), 'view-fullscreen',
 			shct=Qt.Key_F11, trigbool=self.enableFullScreen)
+		self.actionFullScreen.setChecked(self.isFullScreen())
 		self.actionFullScreen.setPriority(QAction.LowPriority)
 		self.actionConfig = self.act(self.tr('Preferences'), icon='preferences-system',
 			trig=self.openConfigDialog)
@@ -169,6 +178,12 @@ class ReTextWindow(QMainWindow):
 			lambda: self.currentTab.editBox.cut(), shct=QKeySequence.Cut)
 		self.actionPaste = self.act(self.tr('Paste'), 'edit-paste',
 			lambda: self.currentTab.editBox.paste(), shct=QKeySequence.Paste)
+		self.actionPasteImage = self.act(self.tr('Paste image'), 'edit-paste',
+			lambda: self.currentTab.editBox.pasteImage(), shct=Qt.CTRL+Qt.SHIFT+Qt.Key_V)
+		self.actionMoveUp = self.act(self.tr('Move line up'), 'go-up',
+			lambda: self.currentTab.editBox.moveLineUp(), shct=Qt.ALT+Qt.Key_Up)
+		self.actionMoveDown = self.act(self.tr('Move line down'), 'go-down',
+			lambda: self.currentTab.editBox.moveLineDown(), shct=Qt.ALT+Qt.Key_Down)
 		self.actionUndo.setEnabled(False)
 		self.actionRedo.setEnabled(False)
 		self.actionCopy.setEnabled(False)
@@ -180,17 +195,29 @@ class ReTextWindow(QMainWindow):
 			self.actionEnableSC = self.act(self.tr('Enable'), trigbool=self.enableSpellCheck)
 			self.actionSetLocale = self.act(self.tr('Set locale'), trig=self.changeLocale)
 		self.actionWebKit = self.act(self.tr('Use WebKit renderer'), trigbool=self.enableWebKit)
-		if ReTextWebPreview is None:
+		if ReTextWebKitPreview is None:
 			globalSettings.useWebKit = False
 			self.actionWebKit.setEnabled(False)
 		self.actionWebKit.setChecked(globalSettings.useWebKit)
+		self.actionWebEngine = self.act(self.tr('Use WebEngine (Chromium) renderer'),
+			trigbool=self.enableWebEngine)
+		if ReTextWebEnginePreview is None:
+			globalSettings.useWebEngine = False
+		self.actionWebEngine.setChecked(globalSettings.useWebEngine)
 		self.actionShow = self.act(self.tr('Show directory'), 'system-file-manager', self.showInDir)
 		self.actionFind = self.act(self.tr('Next'), 'go-next', self.find,
 			shct=QKeySequence.FindNext)
 		self.actionFindPrev = self.act(self.tr('Previous'), 'go-previous',
 			lambda: self.find(back=True), shct=QKeySequence.FindPrevious)
+		self.actionReplace = self.act(self.tr('Replace'), 'edit-find-replace',
+			lambda: self.find(replace=True))
+		self.actionReplaceAll = self.act(self.tr('Replace all'), trig=self.replaceAll)
+		menuReplace = QMenu()
+		menuReplace.addAction(self.actionReplaceAll)
+		self.actionReplace.setMenu(menuReplace)
 		self.actionCloseSearch = self.act(self.tr('Close'), 'window-close',
-			lambda: self.searchBar.setVisible(False))
+			lambda: self.searchBar.setVisible(False),
+			shct=QKeySequence.Cancel)
 		self.actionCloseSearch.setPriority(QAction.LowPriority)
 		self.actionHelp = self.act(self.tr('Get help online'), 'help-contents', self.openHelp)
 		self.aboutWindowTitle = self.tr('About ReText')
@@ -202,18 +229,12 @@ class ReTextWindow(QMainWindow):
 		availableMarkups = markups.get_available_markups()
 		if not availableMarkups:
 			print('Warning: no markups are available!')
-		defaultMarkup = availableMarkups[0] if availableMarkups else None
-		if globalSettings.defaultMarkup:
-			mc = markups.find_markup_class_by_name(globalSettings.defaultMarkup)
-			if mc and mc.available():
-				defaultMarkup = mc
-		self.setDefaultMarkup(defaultMarkup)
 		if len(availableMarkups) > 1:
 			self.chooseGroup = QActionGroup(self)
 			markupActions = []
 			for markup in availableMarkups:
 				markupAction = self.act(markup.name, trigbool=self.markupFunction(markup))
-				if markup == self.defaultMarkup:
+				if markup.name == globalSettings.defaultMarkup:
 					markupAction.setChecked(True)
 				self.chooseGroup.addAction(markupAction)
 				markupActions.append(markupAction)
@@ -224,8 +245,9 @@ class ReTextWindow(QMainWindow):
 		self.actionUnderline = self.act(self.tr('Underline'), shct=QKeySequence.Underline,
 			trig=lambda: self.insertFormatting('underline'))
 		self.usefulTags = ('header', 'italic', 'bold', 'underline', 'numbering',
-			'bullets', 'image', 'link', 'inline code', 'code block', 'blockquote')
-		self.usefulChars = ('deg', 'divide', 'dollar', 'hellip', 'laquo', 'larr',
+			'bullets', 'image', 'link', 'inline code', 'code block', 'blockquote',
+			'table')
+		self.usefulChars = ('deg', 'divide', 'euro', 'hellip', 'laquo', 'larr',
 			'lsquo', 'mdash', 'middot', 'minus', 'nbsp', 'ndash', 'raquo',
 			'rarr', 'rsquo', 'times')
 		self.formattingBox = QComboBox(self.editBar)
@@ -254,6 +276,7 @@ class ReTextWindow(QMainWindow):
 		menuFile.addSeparator()
 		menuFile.addAction(self.actionNextTab)
 		menuFile.addAction(self.actionPrevTab)
+		menuFile.addAction(self.actionCloseCurrentTab)
 		menuFile.addSeparator()
 		menuExport = menuFile.addMenu(self.tr('Export'))
 		menuExport.addAction(self.actionSaveHtml)
@@ -274,12 +297,17 @@ class ReTextWindow(QMainWindow):
 		menuEdit.addAction(self.actionCut)
 		menuEdit.addAction(self.actionCopy)
 		menuEdit.addAction(self.actionPaste)
+		menuEdit.addAction(self.actionPasteImage)
+		menuEdit.addSeparator()
+		menuEdit.addAction(self.actionMoveUp)
+		menuEdit.addAction(self.actionMoveDown)
 		menuEdit.addSeparator()
 		if enchant is not None:
 			menuSC = menuEdit.addMenu(self.tr('Spell check'))
 			menuSC.addAction(self.actionEnableSC)
 			menuSC.addAction(self.actionSetLocale)
 		menuEdit.addAction(self.actionSearch)
+		menuEdit.addAction(self.actionGoToLine)
 		menuEdit.addAction(self.actionChangeEditorFont)
 		menuEdit.addAction(self.actionChangePreviewFont)
 		menuEdit.addSeparator()
@@ -291,11 +319,16 @@ class ReTextWindow(QMainWindow):
 		menuFormat.addAction(self.actionBold)
 		menuFormat.addAction(self.actionItalic)
 		menuFormat.addAction(self.actionUnderline)
-		menuEdit.addAction(self.actionWebKit)
+		if ReTextWebKitPreview is not None or ReTextWebEnginePreview is None:
+			menuEdit.addAction(self.actionWebKit)
+		else:
+			menuEdit.addAction(self.actionWebEngine)
 		menuEdit.addSeparator()
 		menuEdit.addAction(self.actionViewHtml)
 		menuEdit.addAction(self.actionPreview)
+		menuEdit.addAction(self.actionInsertTable)
 		menuEdit.addAction(self.actionTableMode)
+		menuEdit.addAction(self.actionInsertImages)
 		if ReTextFakeVimHandler:
 			menuEdit.addAction(self.actionFakeVimMode)
 		menuEdit.addSeparator()
@@ -305,15 +338,15 @@ class ReTextWindow(QMainWindow):
 		menuHelp.addSeparator()
 		menuHelp.addAction(self.actionAbout)
 		menuHelp.addAction(self.actionAboutQt)
-		toolBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
-		toolBar.addAction(self.actionNew)
-		toolBar.addSeparator()
-		toolBar.addAction(self.actionOpen)
-		toolBar.addAction(self.actionSave)
-		toolBar.addAction(self.actionPrint)
-		toolBar.addSeparator()
-		toolBar.addAction(self.actionPreview)
-		toolBar.addAction(self.actionFullScreen)
+		self.toolBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
+		self.toolBar.addAction(self.actionNew)
+		self.toolBar.addSeparator()
+		self.toolBar.addAction(self.actionOpen)
+		self.toolBar.addAction(self.actionSave)
+		self.toolBar.addAction(self.actionPrint)
+		self.toolBar.addSeparator()
+		self.toolBar.addAction(self.actionPreview)
+		self.toolBar.addAction(self.actionFullScreen)
 		self.editBar.addAction(self.actionUndo)
 		self.editBar.addAction(self.actionRedo)
 		self.editBar.addSeparator()
@@ -326,12 +359,17 @@ class ReTextWindow(QMainWindow):
 		self.searchEdit = QLineEdit(self.searchBar)
 		self.searchEdit.setPlaceholderText(self.tr('Search'))
 		self.searchEdit.returnPressed.connect(self.find)
+		self.replaceEdit = QLineEdit(self.searchBar)
+		self.replaceEdit.setPlaceholderText(self.tr('Replace with'))
+		self.replaceEdit.returnPressed.connect(self.find)
 		self.csBox = QCheckBox(self.tr('Case sensitively'), self.searchBar)
 		self.searchBar.addWidget(self.searchEdit)
+		self.searchBar.addWidget(self.replaceEdit)
 		self.searchBar.addSeparator()
 		self.searchBar.addWidget(self.csBox)
 		self.searchBar.addAction(self.actionFindPrev)
 		self.searchBar.addAction(self.actionFind)
+		self.searchBar.addAction(self.actionReplace)
 		self.searchBar.addAction(self.actionCloseSearch)
 		self.searchBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
 		self.searchBar.setVisible(False)
@@ -346,28 +384,33 @@ class ReTextWindow(QMainWindow):
 			try:
 				enchant.Dict(self.sl or None)
 			except enchant.errors.Error as e:
-				print(e, file=sys.stderr)
+				warnings.warn(str(e), RuntimeWarning)
 				globalSettings.spellCheck = False
 			if globalSettings.spellCheck:
 				self.actionEnableSC.setChecked(True)
 		self.fileSystemWatcher = QFileSystemWatcher()
 		self.fileSystemWatcher.fileChanged.connect(self.fileChanged)
 
+	def restoreLastOpenedFiles(self):
+		for file in readListFromSettings("lastFileList"):
+			self.openFileWrapper(file)
+
+		# Show the tab of last opened file
+		lastTabIndex = globalSettings.lastTabIndex
+		if lastTabIndex >= 0 and lastTabIndex < self.tabWidget.count():
+			self.tabWidget.setCurrentIndex(lastTabIndex)
+
 	def iterateTabs(self):
 		for i in range(self.tabWidget.count()):
 			yield self.tabWidget.widget(i)
 
 	def updateStyleSheet(self):
+		self.ss = None
 		if globalSettings.styleSheet:
 			sheetfile = QFile(globalSettings.styleSheet)
 			sheetfile.open(QIODevice.ReadOnly)
 			self.ss = QTextStream(sheetfile).readAll()
 			sheetfile.close()
-		else:
-			palette = QApplication.palette()
-			self.ss = 'html { color: %s; }\n' % palette.color(QPalette.WindowText).name()
-			self.ss += 'td, th { border: 1px solid #c3c3c3; padding: 0 3px 0 3px; }\n'
-			self.ss += 'table { border-collapse: collapse; }\n'
 
 	def initTabWidget(self):
 		def dragEnterEvent(e):
@@ -382,6 +425,7 @@ class ReTextWindow(QMainWindow):
 		self.tabWidget.setMovable(True)
 		self.tabWidget.dragEnterEvent = dragEnterEvent
 		self.tabWidget.dropEvent = dropEvent
+		self.tabWidget.setTabBarAutoHide(globalSettings.tabBarAutoHide)
 
 	def act(self, name, icon=None, trig=None, trigbool=None, shct=None):
 		if not isinstance(shct, QKeySequence):
@@ -400,13 +444,20 @@ class ReTextWindow(QMainWindow):
 		return action
 
 	def actIcon(self, name):
-		return QIcon.fromTheme(name, QIcon(icon_path+name+'.png'))
+		return QIcon.fromTheme(name, QIcon(getBundledIcon(name)))
 
 	def printError(self):
 		import traceback
 		print('Exception occurred while parsing document:', file=sys.stderr)
 		traceback.print_exc()
 
+	def updateTabTitle(self, ind, tab):
+		changed = tab.editBox.document().isModified()
+		if changed and not self.autoSaveActive(tab):
+			title = tab.getBaseName() + '*'
+		else:
+			title = tab.getBaseName()
+		self.tabWidget.setTabText(ind, title)
 
 	def tabFileNameChanged(self, tab):
 		'''
@@ -416,8 +467,10 @@ class ReTextWindow(QMainWindow):
 		if tab == self.currentTab:
 			if tab.fileName:
 				self.setWindowTitle("")
+				if globalSettings.windowTitleFullPath:
+					self.setWindowTitle(tab.fileName + '[*]')
 				self.setWindowFilePath(tab.fileName)
-				self.tabWidget.setTabText(self.ind, tab.getBaseName())
+				self.updateTabTitle(self.ind, tab)
 				self.tabWidget.setTabToolTip(self.ind, tab.fileName)
 				QDir.setCurrent(QFileInfo(tab.fileName).dir().path())
 			else:
@@ -448,21 +501,32 @@ class ReTextWindow(QMainWindow):
 		Perform all UI state changes that need to be done when the
 		modification state of the current tab has changed.
 		'''
+
 		if tab == self.currentTab:
 			changed = tab.editBox.document().isModified()
 			if self.autoSaveActive(tab):
 				changed = False
 			self.actionSave.setEnabled(changed)
+			self.updateTabTitle(self.ind, tab)
 			self.setWindowModified(changed)
 
 	def createTab(self, fileName):
-		self.currentTab = ReTextTab(self, fileName, self.defaultMarkup,
-			previewState=int(globalSettings.livePreviewByDefault))
+		previewStatesByName = {
+			'editor': PreviewDisabled,
+			'normal-preview': PreviewNormal,
+			'live-preview': PreviewLive,
+		}
+		previewState = previewStatesByName.get(globalSettings.defaultPreviewState, PreviewDisabled)
+		if previewState == PreviewNormal and not fileName:
+			previewState = PreviewDisabled  # Opening empty document in preview mode makes no sense
+		self.currentTab = ReTextTab(self, fileName, previewState)
 		self.currentTab.fileNameChanged.connect(lambda: self.tabFileNameChanged(self.currentTab))
 		self.currentTab.modificationStateChanged.connect(lambda: self.tabModificationStateChanged(self.currentTab))
 		self.currentTab.activeMarkupChanged.connect(lambda: self.tabActiveMarkupChanged(self.currentTab))
 		self.tabWidget.addTab(self.currentTab, self.tr("New document"))
 		self.currentTab.updateBoxesVisibility()
+		if previewState > 0:
+			QTimer.singleShot(500, self.currentTab.triggerPreviewUpdate)
 
 	def closeTab(self, ind):
 		if self.maybeSave(ind):
@@ -508,41 +572,48 @@ class ReTextWindow(QMainWindow):
 	def changeEditorFont(self):
 		font, ok = QFontDialog.getFont(globalSettings.editorFont, self)
 		if ok:
-			globalSettings.editorFont = font
-			for tab in self.iterateTabs():
-				tab.editBox.updateFont()
+			self.setEditorFont(font)
+
+	def setEditorFont(self, font):
+		globalSettings.editorFont = font
+		for tab in self.iterateTabs():
+			tab.editBox.updateFont()
 
 	def changePreviewFont(self):
 		font, ok = QFontDialog.getFont(globalSettings.font, self)
 		if ok:
-			globalSettings.font = font
-			for tab in self.iterateTabs():
-				tab.triggerPreviewUpdate()
+			self.setPreviewFont(font)
+
+	def setPreviewFont(self, font):
+		globalSettings.font = font
+		for tab in self.iterateTabs():
+			tab.triggerPreviewUpdate()
 
 	def preview(self, viewmode):
 		self.currentTab.previewState = viewmode * 2
 		self.actionLivePreview.setChecked(False)
 		self.editBar.setDisabled(viewmode)
 		self.currentTab.updateBoxesVisibility()
+		self.currentTab.triggerPreviewUpdate()
 
 	def enableLivePreview(self, livemode):
 		self.currentTab.previewState = int(livemode)
 		self.actionPreview.setChecked(livemode)
 		self.editBar.setEnabled(True)
 		self.currentTab.updateBoxesVisibility()
+		self.currentTab.triggerPreviewUpdate()
 
 	def enableWebKit(self, enable):
 		globalSettings.useWebKit = enable
+		globalSettings.useWebEngine = False
 		for tab in self.iterateTabs():
-			tab.previewBox.disconnectExternalSignals()
-			tab.previewBox.setParent(None)
-			tab.previewBox.deleteLater()
-			tab.previewBox = tab.createPreviewBox(tab.editBox)
-			tab.previewBox.setMinimumWidth(125)
-			tab.addWidget(tab.previewBox)
-			tab.setSizes((50, 50))
-			tab.triggerPreviewUpdate()
-			tab.updateBoxesVisibility()
+			tab.rebuildPreviewBox()
+
+	def enableWebEngine(self, enable):
+		globalSettings.useWebKit = False
+		globalSettings.useWebEngine = enable
+		for tab in self.iterateTabs():
+			tab.rebuildPreviewBox()
 
 	def enableCopy(self, copymode):
 		self.actionCopy.setEnabled(copymode)
@@ -564,7 +635,7 @@ class ReTextWindow(QMainWindow):
 		if yes:
 			FakeVimMode.init(self)
 			for tab in self.iterateTabs():
-				tab.installFakeVimHandler()
+				tab.editBox.installFakeVimHandler()
 		else:
 			FakeVimMode.exit(self)
 
@@ -599,19 +670,34 @@ class ReTextWindow(QMainWindow):
 			if localedlg.checkBox.isChecked():
 				globalSettings.spellCheckLocale = sl
 
+	def search(self):
+		self.searchBar.setVisible(True)
+		self.searchEdit.setFocus(Qt.ShortcutFocusReason)
+
+	def goToLine(self):
+		line, ok = QInputDialog.getInt(self, self.tr("Go to line"), self.tr("Type the line number"))
+		if ok:
+			self.currentTab.goToLine(line-1)
+
 	def searchBarVisibilityChanged(self, visible):
-		self.actionSearch.setChecked(visible)
 		if visible:
 			self.searchEdit.setFocus(Qt.ShortcutFocusReason)
 
-	def find(self, back=False):
+	def find(self, back=False, replace=False):
 		flags = QTextDocument.FindFlags()
 		if back:
 			flags |= QTextDocument.FindBackward
 		if self.csBox.isChecked():
 			flags |= QTextDocument.FindCaseSensitively
 		text = self.searchEdit.text()
-		found = self.currentTab.find(text, flags)
+		replaceText = self.replaceEdit.text() if replace else None
+		found = self.currentTab.find(text, flags, replaceText=replaceText)
+		self.setSearchEditColor(found)
+
+	def replaceAll(self):
+		text = self.searchEdit.text()
+		replaceText = self.replaceEdit.text()
+		found = self.currentTab.replaceAll(text, replaceText)
 		self.setSearchEditColor(found)
 
 	def setSearchEditColor(self, found):
@@ -633,8 +719,9 @@ class ReTextWindow(QMainWindow):
 			if fileName in files:
 				files.remove(fileName)
 			files.insert(0, fileName)
-			if len(files) > 10:
-				del files[10:]
+			recentCount = globalSettings.recentDocumentsCount
+			if len(files) > recentCount:
+				del files[recentCount:]
 			writeListToSettings("recentFileList", files)
 
 	def createNew(self, text=None):
@@ -715,7 +802,7 @@ class ReTextWindow(QMainWindow):
 			if mimetype is None:
 				enabled = True
 			elif markupClass == markups.MarkdownMarkup:
-				enabled = (mimetype in ("text/x-retext-markdown", "text/x-markdown"))
+				enabled = (mimetype in ("text/x-retext-markdown", "text/x-markdown", "text/markdown"))
 			elif markupClass == markups.ReStructuredTextMarkup:
 				enabled = (mimetype in ("text/x-retext-rst", "text/x-rst"))
 			else:
@@ -741,11 +828,12 @@ class ReTextWindow(QMainWindow):
 			supportedExtensions += markup.file_extensions
 		fileFilter = ' (' + str.join(' ', ['*'+ext for ext in supportedExtensions]) + ');;'
 		fileNames = QFileDialog.getOpenFileNames(self,
-			self.tr("Select one or several files to open"), "",
+			self.tr("Select one or several files to open"), QDir.currentPath(),
 			self.tr("Supported files") + fileFilter + self.tr("All files (*)"))
 		for fileName in fileNames[0]:
 			self.openFileWrapper(fileName)
 
+	@pyqtSlot(str)
 	def openFileWrapper(self, fileName):
 		if not fileName:
 			return
@@ -790,13 +878,15 @@ class ReTextWindow(QMainWindow):
 
 	def saveAll(self):
 		for tab in self.iterateTabs():
-			if tab.fileName and QFileInfo(tab.fileName).isWritable():
+			if (tab.fileName and tab.editBox.document().isModified()
+				and QFileInfo(tab.fileName).isWritable()):
 				tab.saveTextToFile()
 
 	def saveFile(self, dlg=False):
 		fileNameToSave = self.currentTab.fileName
 
 		if (not fileNameToSave) or dlg:
+			proposedFileName = ""
 			markupClass = self.currentTab.getActiveMarkupClass()
 			if (markupClass is None) or not hasattr(markupClass, 'default_extension'):
 				defaultExt = self.tr("Plain text (*.txt)")
@@ -812,11 +902,19 @@ class ReTextWindow(QMainWindow):
 					ext = globalSettings.restDefaultFileExtension
 				else:
 					ext = markupClass.default_extension
+			if fileNameToSave is not None:
+				proposedFileName = fileNameToSave
 			fileNameToSave = QFileDialog.getSaveFileName(self,
-				self.tr("Save file"), "", defaultExt)[0]
+				self.tr("Save file"), proposedFileName, defaultExt)[0]
 			if fileNameToSave:
 				if not QFileInfo(fileNameToSave).suffix():
 					fileNameToSave += ext
+				# Make sure we don't overwrite a file opened in other tab
+				for tab in self.iterateTabs():
+					if tab is not self.currentTab and tab.fileName == fileNameToSave:
+						QMessageBox.warning(self, "",
+							self.tr("Cannot save to file which is open in another tab!"))
+						return False
 				self.actionSetEncoding.setDisabled(self.autoSaveActive())
 		if fileNameToSave:
 			if self.currentTab.saveTextToFile(fileNameToSave):
@@ -831,12 +929,15 @@ class ReTextWindow(QMainWindow):
 		if not QFileInfo(fileName).suffix():
 			fileName += ".html"
 		try:
-			_, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=False,
-				                                              webenv=True)
+			_, htmltext, _ = self.currentTab.getDocumentForExport(webenv=True)
 		except Exception:
 			return self.printError()
 		htmlFile = QFile(fileName)
-		htmlFile.open(QIODevice.WriteOnly)
+		result = htmlFile.open(QIODevice.WriteOnly)
+		if not result:
+			QMessageBox.warning(self, '',
+				self.tr("Cannot save to file because it is read-only!"))
+			return
 		html = QTextStream(htmlFile)
 		if globalSettings.defaultCodec:
 			html.setCodec(globalSettings.defaultCodec)
@@ -846,21 +947,18 @@ class ReTextWindow(QMainWindow):
 	def textDocument(self, title, htmltext):
 		td = QTextDocument()
 		td.setMetaInformation(QTextDocument.DocumentTitle, title)
-		if self.ss:
-			td.setDefaultStyleSheet(self.ss)
 		td.setHtml(htmltext)
 		td.setDefaultFont(globalSettings.font)
 		return td
 
 	def saveOdf(self):
-		title, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=True,
-		                                                          webenv=False)
+		title, htmltext, _ = self.currentTab.getDocumentForExport()
 		try:
 			document = self.textDocument(title, htmltext)
 		except Exception:
 			return self.printError()
 		fileName = QFileDialog.getSaveFileName(self,
-			self.tr("Export document to ODT"), "",
+			self.tr("Export document to ODT"), self.currentTab.getBaseName() + ".odt",
 			self.tr("OpenDocument text files (*.odt)"))[0]
 		if not QFileInfo(fileName).suffix():
 			fileName += ".odt"
@@ -870,7 +968,7 @@ class ReTextWindow(QMainWindow):
 
 	def saveFileHtml(self):
 		fileName = QFileDialog.getSaveFileName(self,
-			self.tr("Save file"), "",
+			self.tr("Save file"), self.currentTab.getBaseName() + ".html",
 			self.tr("HTML files (*.html *.htm)"))[0]
 		if fileName:
 			self.saveHtml(fileName)
@@ -887,17 +985,53 @@ class ReTextWindow(QMainWindow):
 		printer = QPrinter(QPrinter.HighResolution)
 		printer.setDocName(title)
 		printer.setCreator('ReText %s' % app_version)
+		if globalSettings.paperSize:
+			pageSize = self.getPageSizeByName(globalSettings.paperSize)
+			if pageSize is not None:
+				printer.setPaperSize(pageSize)
+			else:
+				QMessageBox.warning(self, '',
+					self.tr('Unrecognized paperSize setting "%s".') %
+					globalSettings.paperSize)
 		return printer
 
+	def getPageSizeByName(self, pageSizeName):
+		""" Returns a validated PageSize instance corresponding to the given
+		name. Returns None if the name is not a valid PageSize.
+		"""
+		pageSize = None
+
+		lowerCaseNames = {pageSize.lower(): pageSize for pageSize in
+		                  self.availablePageSizes()}
+		if pageSizeName.lower() in lowerCaseNames:
+			pageSize = getattr(QPagedPaintDevice, lowerCaseNames[pageSizeName.lower()])
+
+		return pageSize
+
+	def availablePageSizes(self):
+		""" List available page sizes. """
+
+		sizes = [x for x in dir(QPagedPaintDevice)
+		         if type(getattr(QPagedPaintDevice, x)) == QPagedPaintDevice.PageSize]
+		return sizes
+
 	def savePdf(self):
 		fileName = QFileDialog.getSaveFileName(self,
 			self.tr("Export document to PDF"),
-			"", self.tr("PDF files (*.pdf)"))[0]
+			self.currentTab.getBaseName() + ".pdf",
+			self.tr("PDF files (*.pdf)"))[0]
 		if fileName:
 			if not QFileInfo(fileName).suffix():
 				fileName += ".pdf"
-			title, htmltext, preview = self.currentTab.getDocumentForExport(includeStyleSheet=True,
-										        webenv=False)
+			title, htmltext, preview = self.currentTab.getDocumentForExport()
+			if globalSettings.useWebEngine and hasattr(preview.page(), "printToPdf"):
+				pageSize = self.getPageSizeByName(globalSettings.paperSize)
+				if pageSize is None:
+					pageSize = QPageSize(QPageSize.A4)
+				margins = QMarginsF(20, 20, 13, 20)  # left, top, right, bottom (in millimeters)
+				layout = QPageLayout(pageSize, QPageLayout.Portrait, margins, QPageLayout.Millimeter)
+				preview.page().printToPdf(fileName, layout)  # Available since Qt 5.7
+				return
 			printer = self.standardPrinter(title)
 			printer.setOutputFormat(QPrinter.PdfFormat)
 			printer.setOutputFileName(fileName)
@@ -906,8 +1040,7 @@ class ReTextWindow(QMainWindow):
 				document.print(printer)
 
 	def printFile(self):
-		title, htmltext, preview = self.currentTab.getDocumentForExport(includeStyleSheet=True,
-										webenv=False)
+		title, htmltext, preview = self.currentTab.getDocumentForExport()
 		printer = self.standardPrinter(title)
 		dlg = QPrintDialog(printer, self)
 		dlg.setWindowTitle(self.tr("Print document"))
@@ -917,8 +1050,7 @@ class ReTextWindow(QMainWindow):
 				document.print(printer)
 
 	def printPreview(self):
-		title, htmltext, preview = self.currentTab.getDocumentForExport(includeStyleSheet=True,
-										webenv=False)
+		title, htmltext, preview = self.currentTab.getDocumentForExport()
 		document = self.getDocumentForPrint(title, htmltext, preview)
 		if document is None:
 			return
@@ -967,9 +1099,16 @@ class ReTextWindow(QMainWindow):
 	def clipboardDataChanged(self):
 		mimeData = QApplication.instance().clipboard().mimeData()
 		if mimeData is not None:
-			self.actionPaste.setEnabled(mimeData.hasText() or mimeData.hasImage())
+			self.actionPaste.setEnabled(mimeData.hasText())
+			self.actionPasteImage.setEnabled(mimeData.hasImage())
 
 	def insertFormatting(self, formatting):
+		if formatting == 'table':
+			dialog = InsertTableDialog(self)
+			dialog.show()
+			self.formattingBox.setCurrentIndex(0)
+			return
+
 		cursor = self.currentTab.editBox.textCursor()
 		text = cursor.selectedText()
 		moveCursorTo = None
@@ -1021,22 +1160,24 @@ class ReTextWindow(QMainWindow):
 		self.symbolBox.setCurrentIndex(0)
 
 	def fileChanged(self, fileName):
-		ind = None
-		for testind, tab in enumerate(self.iterateTabs()):
-			if tab.fileName == fileName:
-				ind = testind
-		if ind is None:
+		tab = None
+		for testtab in self.iterateTabs():
+			if testtab.fileName == fileName:
+				tab = testtab
+		if tab is None:
 			self.fileSystemWatcher.removePath(fileName)
-		self.tabWidget.setCurrentIndex(ind)
+			return
 		if not QFile.exists(fileName):
-			self.currentTab.editBox.document().setModified(True)
+			self.tabWidget.setCurrentWidget(tab)
+			tab.editBox.document().setModified(True)
 			QMessageBox.warning(self, '', self.tr(
 				'This file has been deleted by other application.\n'
 				'Please make sure you save the file before exit.'))
-		elif not self.currentTab.editBox.document().isModified():
+		elif not tab.editBox.document().isModified():
 			# File was not modified in ReText, reload silently
-			self.currentTab.readTextFromFile()
+			tab.readTextFromFile()
 		else:
+			self.tabWidget.setCurrentWidget(tab)
 			text = self.tr(
 				'This document has been modified by other application.\n'
 				'Do you want to reload the file (this will discard all '
@@ -1050,10 +1191,10 @@ class ReTextWindow(QMainWindow):
 			messageBox.addButton(QMessageBox.Cancel)
 			messageBox.exec()
 			if messageBox.clickedButton() is reloadButton:
-				self.currentTab.readTextFromFile()
+				tab.readTextFromFile()
 			else:
 				self.autoSaveEnabled = False
-				self.currentTab.editBox.document().setModified(True)
+				tab.editBox.document().setModified(True)
 		if fileName not in self.fileSystemWatcher.files():
 			# https://github.com/retext-project/retext/issues/137
 			self.fileSystemWatcher.addPath(fileName)
@@ -1079,15 +1220,18 @@ class ReTextWindow(QMainWindow):
 		for ind in range(self.tabWidget.count()):
 			if not self.maybeSave(ind):
 				return closeevent.ignore()
-		if globalSettings.saveWindowGeometry and not self.isMaximized():
+		if globalSettings.saveWindowGeometry:
 			globalSettings.windowGeometry = self.saveGeometry()
+		if globalSettings.openLastFilesOnStartup:
+			files = [tab.fileName for tab in self.iterateTabs()]
+			writeListToSettings("lastFileList", files)
+			globalSettings.lastTabIndex = self.tabWidget.currentIndex()
 		closeevent.accept()
 
 	def viewHtml(self):
 		htmlDlg = HtmlDialog(self)
 		try:
-			_, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=False,
-			                                                      webenv=False)
+			_, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=False)
 		except Exception:
 			return self.printError()
 		winTitle = self.currentTab.getBaseName()
@@ -1098,6 +1242,23 @@ class ReTextWindow(QMainWindow):
 		htmlDlg.raise_()
 		htmlDlg.activateWindow()
 
+	def insertImages(self):
+		supportedExtensions = ['.png', '.jpg', '.jpeg', '.gif', '.bmp']
+		fileFilter = ' (%s);;' % ' '.join('*' + ext for ext in supportedExtensions)
+		fileNames, _selectedFilter = QFileDialog.getOpenFileNames(self,
+			self.tr("Select one or several images to open"), QDir.currentPath(),
+			self.tr("Supported files") + fileFilter + self.tr("All files (*)"))
+
+		cursor = self.currentTab.editBox.textCursor()
+
+		imagesMarkup = '\n'.join(
+			self.currentTab.editBox.getImageMarkup(fileName)
+			for fileName in fileNames)
+		cursor.insertText(imagesMarkup)
+
+		self.formattingBox.setCurrentIndex(0)
+		self.currentTab.editBox.setFocus(Qt.OtherFocusReason)
+
 	def openHelp(self):
 		QDesktopServices.openUrl(QUrl('https://github.com/retext-project/retext/wiki'))
 
@@ -1106,7 +1267,7 @@ class ReTextWindow(QMainWindow):
 		'<p><b>' + (self.tr('ReText %s (using PyMarkups %s)') % (app_version, markups.__version__))
 		+'</b></p>' + self.tr('Simple but powerful editor'
 		' for Markdown and reStructuredText')
-		+'</p><p>'+self.tr('Author: Dmitry Shachnev, 2011').replace('2011', '2011–2016')
+		+'</p><p>'+self.tr('Author: Dmitry Shachnev, 2011').replace('2011', '2011–2020')
 		+'<br><a href="https://github.com/retext-project/retext">'+self.tr('Website')
 		+'</a> | <a href="http://daringfireball.net/projects/markdown/syntax">'
 		+self.tr('Markdown syntax')
@@ -1114,9 +1275,7 @@ class ReTextWindow(QMainWindow):
 		+self.tr('reStructuredText syntax')+'</a></p>')
 
 	def setDefaultMarkup(self, markupClass):
-		self.defaultMarkup = markupClass
-		defaultName = markups.get_available_markups()[0].name
-		writeToSettings('defaultMarkup', markupClass.name, defaultName)
+		globalSettings.defaultMarkup = markupClass.name
 		for tab in self.iterateTabs():
 			if not tab.fileName:
-				tab.setDefaultMarkupClass(markupClass)
+				tab.updateActiveMarkupClass()
